@@ -8,29 +8,17 @@
 # just a few useful routines for my current project.
 
 from ioctl import Structure, t
+from util import pretty, humanize
+
 import collections
 import ioctl
 import logging
 import os.path
-import pprint
+
 logger = logging.getLogger(__name__)
 # logger.setLevel('DEBUG')
 
 
-def pretty(obj):
-    """ Return pretty representation of obj. """
-    # if True:
-    #     return pprint.pformat(dict(obj.__dict__))
-    return pprint.pformat(obj)
-    # try:
-    #     return pprint.pformat(dict(obj))
-    # except TypeError:
-    #     try:
-    #         return pprint.pformat(obj.__dict__)
-    #     except KeyError:
-    #         logger.exception("Funny error.")
-
-
 def bytes2uuid(b):
     """ Return standard human-friendly UUID. """
     if b.strip(chr(0)) == '':
@@ -186,10 +174,10 @@ btrfs_root_item = Structure(
     (t.u8, 'uuid', BTRFS_UUID_SIZE, bytes2uuid, uuid2bytes),
     (t.u8, 'parent_uuid', BTRFS_UUID_SIZE, bytes2uuid, uuid2bytes),
     (t.u8, 'received_uuid', BTRFS_UUID_SIZE, bytes2uuid, uuid2bytes),
-    (t.le64, 'ctransid'),
-    (t.le64, 'otransid'),
-    (t.le64, 'stransid'),
-    (t.le64, 'rtransid'),
+    (t.le64, 'ctransid'),  # updated when an inode changes
+    (t.le64, 'otransid'),  # trans when created
+    (t.le64, 'stransid'),  # trans when sent. non-zero for received subvol
+    (t.le64, 'rtransid'),  # trans when received. non-zero for received subvol
     (btrfs_timespec, 'ctime'),
     (btrfs_timespec, 'otime'),
     (btrfs_timespec, 'stime'),
@@ -279,6 +267,23 @@ btrfs_ioctl_vol_args_v2 = Structure(
     packed=True
 )
 
+btrfs_ioctl_timespec = Structure(
+    (t.u64, 'sec'),
+    (t.u32, 'nsec'),
+    packed=True
+)
+
+btrfs_ioctl_received_subvol_args = Structure(
+    (t.char, 'uuid', BTRFS_UUID_SIZE),   # /* in */
+    (t.u64, 'stransid'),                # /* in */
+    (t.u64, 'rtransid'),                # /* out */
+    (btrfs_ioctl_timespec, 'stime'),        # /* in */
+    (btrfs_ioctl_timespec, 'rtime'),        # /* out */
+    (t.u64, 'flags'),                   # /* in */
+    (t.u64, 'reserved', 16),             # /* in */
+    packed=True
+)
+
 BTRFS_IOCTL_MAGIC = 0x94
 
 objectTypeKeys = {
@@ -344,16 +349,17 @@ class Volume(object):
 
     def __init__(self, rootid, generation, info):
         """ Initialize. """
-        # logger.debug("Volume %d: %s", id, pretty(info))
+        logger.debug("Volume %d/%d: %s", rootid, generation, pretty(info))
         self.id = rootid  # id in BTRFS_ROOT_TREE_OBJECTID, also FS treeid for this volume
-        self.original_gen = generation
-        self.current_gen = info.generation
+        self.original_gen = info.otransid
+        self.current_gen = info.ctransid
         # self.size = info.bytes_used
         self.readOnly = bool(info.flags & BTRFS_ROOT_SUBVOL_RDONLY)
         self.level = info.level
         self.uuid = info.uuid
         self.parent_uuid = info.parent_uuid
         self.received_uuid = info.received_uuid
+        self.sent_gen = info.stransid
 
         self.totalSize = None
         self.exclusiveSize = None
@@ -406,8 +412,8 @@ class Volume(object):
         """ String representation. """
         # logger.debug("%d %d %d", self.gen, self.info.generation, self.info.inode.generation)
         # logger.debug("%o %o", self.info.flags, self.info.inode.flags)
-        return """%4d '%s' (level:%d gen:%d total:%d exclusive:%d%s)
-        %s (parent:%s received:%s)
+        return """%4d '%s' (level:%d gen:%d total:%s exclusive:%s%s) %s
+        (parent:%s/%d received:%s/%d)
         %s%s""" % (
             self.id or -1,
             # ", ".join([dirPath + name for (dirPath, name) in self.links.values()]),
@@ -415,12 +421,12 @@ class Volume(object):
             self.level or -1,
             self.current_gen or -1,
             # self.size,
-            self.totalSize or -1,
-            self.exclusiveSize or -1,
+            humanize(self.totalSize or -1),
+            humanize(self.exclusiveSize or -1),
             " ro" if self.readOnly else "",
             self.uuid,
-            self.parent_uuid,
-            self.received_uuid,
+            self.parent_uuid, self.original_gen,
+            self.received_uuid, self.sent_gen,
             "\n\t".join(self.linuxPaths),
             # "\n\t" + pretty(self.__dict__),
             "",
@@ -434,17 +440,17 @@ class Volume(object):
 
     def copy(self, path):
         """ Make another snapshot of this into dirName. """
-        dest = _Directory(os.path.dirname(path))
-        with self._directory() as source:
+        destDir = _Directory(os.path.dirname(path))
+        with self._snapshot() as source, destDir as dest:
             dest.SNAP_CREATE_V2(
                 flags=BTRFS_SUBVOL_RDONLY,
-                name=os.path.basename(path),
+                name=str(os.path.basename(path)),
                 fd=source.fd,
             )
 
-    def _directory(self):
+    def _snapshot(self):
         path = next(iter(self.linuxPaths))
-        return _Directory(path)
+        return SnapShot(path)
 
 
 class Control(ioctl.Control):
@@ -500,8 +506,13 @@ class FileSystem(ioctl.Device):
     def _getDevices(self):
         self.devices = []
         fs = self.FS_INFO()
-        for i in xrange(1, fs.num_devices + 1):
-            dev = self.DEV_INFO(devid=i)
+        for i in xrange(1, fs.max_id + 1):
+            try:
+                dev = self.DEV_INFO(devid=i)
+            except IOError as error:
+                if error.errno == 19:
+                    continue
+                raise
             self.devices.append(dev.path)
 
     def _getMounts(self):
@@ -586,16 +597,15 @@ class FileSystem(ioctl.Device):
                 break
 
             for _ in xrange(results):
-                assert buf.len >= btrfs_ioctl_search_header.size, buf.len
+                # assert buf.len >= btrfs_ioctl_search_header.size, buf.len
                 data = buf.read(btrfs_ioctl_search_header)
 
                 # logger.debug("Object %d: %s", i, pretty(data))
 
-                assert buf.len >= data.len, (buf.len, data.len)
-                yield (data, buf.view(data.len))
+                # assert buf.len >= data.len, (buf.len, data.len)
+                yield (data, buf.readView(data.len))
 
                 key = FileSystem.Key(data.objectid, data.type, data.offset).next()
-                buf.skip(data.len)
 
     def _getRoots(self):
         for (header, buf) in self._walkTree(BTRFS_ROOT_TREE_OBJECTID):
@@ -710,5 +720,13 @@ class _Directory(ioctl.Device):
     SNAP_DESTROY = Control.IOW(15, btrfs_ioctl_vol_args)
     SNAP_CREATE_V2 = Control.IOW(23, btrfs_ioctl_vol_args_v2)
 
+
+class SnapShot(ioctl.Device):
+
+    """ SnapShot (read-only subvolume) identified by path. """
+
+    SET_RECEIVED_SUBVOL = Control.IOWR(37, btrfs_ioctl_received_subvol_args)
+
+
 # define BTRFS_IOC_START_SYNC _IOR(BTRFS_IOCTL_MAGIC, 24, __u64)
 # define BTRFS_IOC_WAIT_SYNC  _IOW(BTRFS_IOCTL_MAGIC, 22, __u64)
